#!/usr/bin/env ruby
# Copyright (C) 2004  Christian Neukirchen <chneukirchen@yahoo.de>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#

# nukumi -- jap. for warmth
$VERBOSE = true

$:.unshift "lib"


require 'time'
require 'find'
require 'optparse'
require 'ftools'

require 'actions'
require 'views'
require 'flavors'
require 'formatters'
require 'utils'


class Nukumi
  VERSION = "0.5"

  begin
    Dir.chdir(File.dirname($0)) {
      rev = `which tla >/dev/null && tla logs -r 2>/dev/null | head -1`.strip
      unless rev == ""
        REVISION = "-" + rev
      else
        REVISION = ""
      end
    }
  rescue
    REVISION = "-" + "$Arch-Revision: patch-57 $".split[1]
  end

  Entry = Struct.new(:header, :body, :time, :ctime, :format, :keys)

  def initialize
    @indir = nil
    @outdir = ""

    @flavors = []
    @views = []
    @handler = []

    @view = {}
    @started = false

    @config = {}
  end

  def entries(indir)
    @indir = indir
  end

  def prefix(prefix=nil)
    if prefix
      @outdir = prefix
      @outdir << "/"  if prefix !~ %r:/$:
    end
    @outdir
  end

  def config(cfg)
    @config.merge! cfg
  end

  def option(name, value)
    @config[name] = value
  end

  def resolve(v, root)
    case v
    when Class
      v.new
    when Symbol
      root.const_get(v).new
    else
      v
    end
  end

  def flavor(f, use=[Object])
    use = [*use]
    use.map! { |v| v.kind_of?(Symbol) ? Nukumi::View.const_get(v) : v }
    @flavors << [resolve(f, Nukumi::Flavor), use]
  end

  def view(v)
    @views << resolve(v, Nukumi::View)
  end
  
  def handler(h)
    @handler << resolve(h, Object)
  end

  def get_entry_list
    @entry_files = []

## Find.find doesn't follow symlinks!
#    Find.find(@dir) do |path|
#      if File.stat(path).file? && path !~ /~$/
#        @entry_files << path
#      end
#    end

    @entry_files = `find #{@indir} -follow -path '*~' -prune \
                                      -o -path '*/.*' -prune \
                                      -o -type f -print0`.split("\0")
  end

  def parse_entry(entry_file)
    entry = Entry.new
    File.open(entry_file) { |file|
      state = :header
      
      entry.header = {}
      entry.body = ""
      
      file.each { |line|
        line.chomp!
        
        if line =~ /^(\w*?): (.*)/
          entry.header[$1] = $2.strip
        elsif line =~ /^$/
          entry.body = file.read
          break
        else
          abort "Misformed entry `#{entry_file}': " +
                "Invalid header line `#{line}'"
        end
      }
      entry.ctime = file.ctime
    }
    
    entry
  end

  def read_entries
    @entries = []

    @entry_files.each { |entry_file|
      entry = parse_entry entry_file

      if entry.header["Subject"].nil?
        abort "Misformed entry `#{entry_file}': No Subject given"
      end

      # Use ctime if no date given.
      entry.header["Date"] ||= entry.ctime.rfc822
      entry.time = Time.parse(entry.header["Date"]).gmtime

      # xmlschema dates are needed for e.g. Atom output.
      entry.header["XDate"] = entry.time.xmlschema

      # Use ctime as modification time.
      entry.header["Modified"] = entry.ctime.gmtime.xmlschema
      
      # Make a unique, linkable timestamp
      entry.header["Permalink"] = "x-" + entry.time.strftime("%Y%m%d-%H%M%S")
      
      entry.header["Directory"] = File.dirname(entry_file.gsub(/^.*?\//, '/'))
      
      entry.header["Encoding"] ||= "BlueCloth"
      entry.format = lookup_formatter entry.header["Encoding"]

      entry.keys = {}
      
      @entries << entry
    }
    
    @entries = @entries.sort_by { |e| e.time }.reverse
  end

  def lookup_formatter(name)
    begin
      require name.downcase
    rescue LoadError
      true
    end

    if Formatters.const_defined? name
      Formatters.const_get name
    elsif Object.const_defined? name
      Object.const_get name
    else
      abort "No such encoding: #{name}"
    end
  end

  def make_views
    @entries.each_with_index { |entry, index|
      @views.each { |view|
        key = view.generate_key(entry, index)
        
        view.keys << key
        entry.keys[view.name] = key || ''
      }
    }

    # Normalize view keys.
    @views.each { |view|
      view.keys = view.keys.compact.uniq.sort.reverse
    }
  end
    
  def generate
    get_entry_list
    read_entries

    make_views

    @views.each { |view|
      @flavors.each { |flavor, use|
        # Is any class to generate a superclass of the flavor (or it
        # itself)?
        if use.any? { |u| view.class <= u }
          generate_view_with_flavor view, flavor
        end
      }
    }
  end

  def generate_view_with_flavor(view, flavor)
    view.keys.flatten.each { |k|
      i = -1
      entries = @entries.find_all { |e|
        i += 1
        [*view.generate_key(e, i)].any? { |nk| nk == k }
      }

      output = @outdir + k + "." + flavor.class::EXTENSION
      
      new = false

      if !(File.exist? output) || entries.any? { |e|
                                    e.ctime > File.ctime(output)
                                  }
        entries.sort!(&view.method(:cmp))

        File.makedirs File.dirname(output)
        flavor.generate entries, output, k, view, @views, @config
        new = true
      end

      @handler.each { |h| h.file output, new }
    }
  end
    
end


if $0 == __FILE__
  rc = "nukumirc"
  begin
    OptionParser.new { |opt|
      opt.banner = "Usage: nukumi.rb [OPTION]... [ACTION]"
      opt.separator ""
      opt.separator "Options:"
      opt.on("-r", "--rc-file RC-FILE",
             "Use the rc-file RC-FILE instead of #{rc}") { |rcfile|
        rc = rcfile
      }
      opt.separator ""
      opt.separator "Common options:"
      opt.on_tail("-h", "--help", "Show this message") do
        puts opt
        exit
      end
      opt.on_tail("--version", "Show version") do
        puts Nukumi::VERSION + Nukumi::REVISION
        exit
      end
    }.parse! ARGV
  rescue OptionParser::ParseError => pe
    abort pe.to_s
  end

  action = ARGV[0] || :generate

  code = File.read(rc)
  n = Nukumi.new
  n.instance_eval code, rc
  action_handler = Nukumi::Actions.new(n, ARGV[2..-1])
  if action_handler.respond_to? action
    action_handler.send action
  else
    abort "No such action: #{action}"
  end
end

# arch-tag: 51245de1-0b70-4611-96a3-959ad46df365
